unit cairoobjects;

{$mode objfpc}{$H+}{$F+}

interface

uses
  Classes, SysUtils, cairo, Graphics, types, Printers, FPCanvas, GraphType,
  Math, Printer4Lazarus;

type
  PPrinter = ^TPrinter;
  TCairoSize = record
    cx: Double;
    cy: Double;
  end;

  TCairoColor = record
    R: Double;
    G: Double;
    B: Double;
    A: Double;
  end;

  TCairoFont = record
    Name: PChar;
    Style: cairo_font_slant_t;
    Weight: cairo_font_weight_t;
    Size: Double;
    Color: TCairoColor;
  end;
  
  TDashes = Array [0..20] of Double;
  
  TCairoDashRec = record
    Dashes: TDashes;
    NumDashes: Integer;
    Offset: Double;
  end;
  
  { TCairoDash }

  TCairoDash = class
    private
      fDashes: TDashes;
      fNumDashes: Integer;
      fOffset: Double;
    protected
      procedure SetDash(ADash: TCairoDashRec);
    public
      PofDashes: ^TDashes;
      constructor Create;
      property Dash: TCairoDashRec write SetDash;
      property NumDashes: Integer read fNumDashes write fNumDashes;
      property Offset: Double read fOffset write fOffset;
  end;
  
  { TCairoPen }

  TCairoPen = class
    private
      fWidth: Double;
      fColor: TCairoColor;
      fDash: TCairoDash;
      fStyle: TFPPenStyle;
    protected
      procedure SetWidth(AWidth: Double);
      procedure SetColor(AColor: TCairoColor);
      procedure SetColor(AColor: TColor);
      procedure SetStyle(AStyle: TFPPenStyle);
    public
      constructor Create;
      property Width: Double read fWidth write SetWidth;
      property Color: TCairoColor read fColor write SetColor;
      property Dash: TCairoDash read fDash write fDash;
      property Style: TFPPenStyle read fStyle write SetStyle default psSolid;
  end;


  TCairoObject = class;

  { TCairoCanvas }

  TCairoCanvas = class
    private
      fOwner: TCairoObject;
      fPen: TCairoPen;
      fBrush: TBrush;
      fFont: TCairoFont;
    protected
      procedure CairoFill;
      procedure CairoStroke;
      procedure SetCairoColor(AColor: TColor);
    public
      constructor Create(AOwner: TCairoObject);
      destructor Destroy;
      procedure Arc(ALeft, ATop, ARight, ABottom, angle1, angle2: Integer); virtual;
      procedure Arc(ALeft, ATop, ARight, ABottom, SX, SY, EX, EY: Integer); virtual;
      //Procedure BrushCopy(Dest: TRect; InternalImages: TBitmap; Src: TRect;
      //                    TransparentColor: TColor); virtual;
      procedure Chord(x1, y1, x2, y2,
                      StartAngle16Deg, EndAngle16Deg: Integer); virtual;
      procedure Chord(x1, y1, x2, y2, SX, SY, EX, EY: Integer); virtual;
      procedure CopyRect(const Dest: TRect; SrcCanvas: TCanvas;
                         const Source: TRect); virtual;
      procedure Draw(X,Y: Integer; SrcGraphic: TGraphic); virtual;
      procedure DrawFocusRect(const ARect: TRect); virtual;
      procedure StretchDraw(const DestRect: TRect; SrcGraphic: TGraphic); virtual;
      procedure Ellipse(const ARect: TRect); // already in fpcanvas
      procedure Ellipse(x1, y1, x2, y2: Integer); virtual; // already in fpcanvas
      procedure FillRect(const ARect: TRect); virtual;
      procedure FillRect(X1,Y1,X2,Y2: Integer);
      procedure FloodFill(X, Y: Integer; FillColor: TColor;
                          FillStyle: TFillStyle); virtual;
      procedure Frame3d(var ARect: TRect; const FrameWidth: integer;
                        const Style: TGraphicsBevelCut); virtual;
      procedure Frame(const ARect: TRect); virtual; // border using pen
      procedure Frame(X1,Y1,X2,Y2: Integer);     // border using pen
      procedure FrameRect(const ARect: TRect); virtual; // border using brush
      procedure FrameRect(X1,Y1,X2,Y2: Integer); // border using brush
      procedure GradientFill(ARect: TRect; AStart, AStop: TColor; ADirection: TGradientDirection);
      procedure Line(X1,Y1,X2,Y2: Integer); virtual; // short for MoveTo();LineTo(); // already in fpcanvas
      procedure Line(const p1,p2: TPoint);
      procedure Line(const Points: TRect);
      procedure LineTo(X1,Y1: Integer); virtual; // already in fpcanvas
      procedure MoveTo(X1,Y1: Integer); virtual; // already in fpcanvas
      procedure RadialPie(x1, y1, x2, y2,
                          StartAngle16Deg, EndAngle16Deg: Integer); virtual;
      procedure Pie(EllipseX1,EllipseY1,EllipseX2,EllipseY2,
                    StartX,StartY,EndX,EndY: Integer); virtual;
      procedure PolyBezier(Points: PPoint; NumPts: Integer;
                           Filled: boolean = False;
                           Continuous: boolean = False); virtual;
      procedure PolyBezier(const Points: array of TPoint;
                           Filled: boolean = False;
                           Continuous: boolean = False);
      procedure Polygon(const Points: array of TPoint;
                        Winding: Boolean;
                        StartIndex: Integer = 0;
                        NumPts: Integer = -1);
      procedure Polygon(Points: PPoint; NumPts: Integer;
                        Winding: boolean = False); virtual;
      procedure Polygon(const Points: array of TPoint); // already in fpcanvas
      procedure Polyline(const Points: array of TPoint;
                         StartIndex: Integer;
                         NumPts: Integer = -1);
      procedure Polyline(Points: PPoint; NumPts: Integer); virtual;
      procedure Polyline(const Points: array of TPoint); // already in fpcanvas
      Procedure Rectangle(X1,Y1,X2,Y2: Integer); virtual; // already in fpcanvas
      Procedure Rectangle(const ARect: TRect); // already in fpcanvas
      Procedure RoundRect(X1, Y1, X2, Y2: Integer; RX,RY: Integer); virtual;
      Procedure RoundRect(const Rect: TRect; RX,RY: Integer);
      procedure TextOut(X,Y: Integer; const Text: String); virtual; // already in fpcanvas
      procedure TextRect(const ARect: TRect; X, Y: integer; const Text: string);
      procedure TextRect(ARect: TRect; X, Y: integer; const Text: string;
                         const Style: TTextStyle); virtual;
      function TextExtent(const Text: string): TCairoSize; virtual;
      function TextHeight(const Text: string): Double; virtual;
      function TextWidth(const Text: string): Double; virtual;
      property Pen: TCairoPen read fPen Write fPen;
      property Brush: TBrush read fBrush write fBrush;
      property Font: TCairoFont read fFont write fFont;
      property Owner: TCairoObject read fOwner;
  end;
  
{ TCairoObject }

TCairoObject = class
  private
    fCanvas: TCairoCanvas;
    fSurface: PCairo_surface_t;
    fCairo: PCairo_t;
  protected
    procedure CreateSurface;
    procedure CreateCairo; virtual;
  public
    constructor Create;
    destructor Destroy;
    procedure BeginDoc;
    procedure EndDoc;
    procedure NewPage;
    property Surface: PCairo_surface_t read fSurface;
    property Cairo: PCairo_t read fCairo;
    property Canvas: TCairoCanvas read fCanvas write fCanvas;
end;

{ TCairoPrinter }

TCairoPrinter = class(TCairoObject)
  private
  protected
    wf: cairo_write_func_t;
    fPrinter: TPrinter;
    procedure CreateSurface;
  public
    constructor Create(APrinter: TPrinter; Awf: cairo_write_func_t);
    procedure BeginDoc;
    procedure NewPage;
    procedure EndDoc;
end;

function WritetoPrinter(closure: pointer; data:PByte; lenght: LongWord):cairo_status_t;cdecl;

var Prn: TCairoPrinter;
implementation


const
  DegreePer16rad = 2*Pi/5760;
  
function CalcAngle(x1, y1, x2, y2: Double): Double;
  var
    x, y, Angle: Double;
  begin
    x := x2 - x1;
    y := y2 - y1;
    if x = 0 then
      Angle := Pi / 2
    else
      Angle := arctan2(y, x);
    if y >= 0 then
      Result := Angle
    else
      Result := Angle + Pi;
  end;
  
function LazColor2CairoColor(Color: TColor): TCairoColor;
  var
    C: TCairoColor;
  begin
    C.R := (Color and $000000FF) / $000000FF;
    C.G := (Color and $0000FF00) / $0000FF00;
    C.B := (Color and $00FF0000) / $00FF0000;
    C.A := (Color and $FF000000) / $FF000000;
    C.A := 1;
    Result := C;
  end;

{ TCairoDash }

procedure TCairoDash.SetDash(ADash: TCairoDashRec);
begin
  fNumDashes := ADash.NumDashes;
  fDashes := ADash.Dashes;
  fOffset := ADash.Offset;
end;

constructor TCairoDash.Create;
begin
  inherited Create;
  fDashes[0] := 0;
  fOffset := 0;
  fNumDashes := 0;
  PofDashes := @fDashes;
end;

{ TCairoPen }

procedure TCairoPen.SetWidth(AWidth: Double);
begin
  fWidth := AWidth;
  SetStyle(fStyle);
end;

procedure TCairoPen.SetColor(AColor: TCairoColor);
begin
  fColor := AColor;
end;

procedure TCairoPen.SetColor(AColor: TColor);
begin
  fColor := LazColor2CairoColor(AColor);
end;

procedure TCairoPen.SetStyle(AStyle: TFPPenStyle);
var
  DashRec: TCairoDashRec;
  i: Integer;
begin
  case AStyle of
    psSolid:
      begin
        with DashRec do
          begin
            Dashes[0] := (0);
            NumDashes := 0;
            Offset := 0;
          end;
          fDash.SetDash(DashRec);
      end;
    psDash:
      begin
        with DashRec do
          begin
            Dashes[0] := (Width*5);
            NumDashes := 1;
            Offset := 0;
          end;
          fDash.SetDash(DashRec);
      end;
    psDot:
      begin
        with DashRec do
          begin
            Dashes[0] := (Width);
            NumDashes := 1;
            Offset := 0;
          end;
          fDash.SetDash(DashRec);
      end;
    psDashDot:
      begin
        with DashRec do
          begin
            Dashes[0] := fWidth*5;
            for i := 1 to 3 do
              Dashes[i] := fWidth;
            NumDashes := 4;
            Offset := 0;
          end;
          fDash.SetDash(DashRec);
      end;
    psDashDotDot:
      begin
        with DashRec do
          begin
            Dashes[0] := fWidth*5;
            for i := 1 to 5 do
              Dashes[i] := fWidth;
            NumDashes := 6;
            Offset := 0;
          end;
          fDash.SetDash(DashRec);
      end;
  end;
end;

constructor TCairoPen.Create;
begin
  inherited Create;
  fDash := TCairoDash.Create;
  Width := 1;
  with Color do
    begin
      A := 1;
      R := 0;
      G := 0;
      B := 0;
    end;
  Style := psDashDot;
end;

{ TCairoCanvas }

procedure TCairoCanvas.CairoFill;
begin
  SetCairoColor(fBrush.Color);
  cairo_fill(Owner.Cairo);
end;

procedure TCairoCanvas.CairoStroke;
begin
  cairo_set_source_rgba(Owner.Cairo, fPen.Color.R, fPen.Color.G, fPen.Color.B, fPen.Color.A);
  cairo_set_dash(Owner.Cairo, PDouble(Pen.Dash.PofDashes), Pen.Dash.NumDashes, Pen.Dash.Offset);
  cairo_set_line_width(Owner.Cairo, fPen.Width);
  if fPen.Width <> 0 then
  cairo_stroke(Owner.Cairo);
end;

procedure TCairoCanvas.SetCairoColor(AColor: TColor);
var
  Color: TCairoColor;
begin
  Color := LazColor2CairoColor(AColor);
  cairo_set_source_rgba(Owner.Cairo, Color.R, Color.G, Color.B, Color.A);
end;

constructor TCairoCanvas.Create(AOwner: TCairoObject);
begin
  inherited Create;
  fPen := TCairoPen.Create;
  fBrush := TBrush.Create;
  fOwner := AOwner;
  fFont.Name := 'serif';
  fFont.Style := CAIRO_FONT_SLANT_NORMAL;
  fFont.Weight := CAIRO_FONT_WEIGHT_NORMAL;
  fFont.Size := 12;
  with fFont.Color do
    begin
      A := 1;
      R := 0;
      G := 0;
      B := 0;
    end;
end;


destructor TCairoCanvas.Destroy;
begin
  fBrush.Destroy;
  fPen.Destroy;
  cairo_destroy(Owner.Cairo);
  inherited Destroy;
end;

procedure TCairoCanvas.Arc(ALeft, ATop, ARight, ABottom, angle1, angle2: Integer
  );
var
  X, Y: Double;

begin
  X := (ARight - ALeft) / 2;
  Y := (ABottom - ATop) / 2;
  cairo_save(Owner.Cairo);
  cairo_translate(Owner.Cairo, ALeft + X, ATop + Y);
  cairo_scale(Owner.Cairo, X, Y);
  cairo_arc(Owner.Cairo, 0, 0, 1, angle1 * DegreePer16Rad, angle2 * DegreePer16Rad);
  cairo_restore(Owner.Cairo);
  CairoStroke;
end;

procedure TCairoCanvas.Arc(ALeft, ATop, ARight, ABottom, SX, SY, EX, EY: Integer
  );
var
  X, Y: Double;
  
begin
  X := (ARight - ALeft) / 2;
  Y := (ABottom - ATop) / 2;
  cairo_save(Owner.Cairo);
  cairo_translate(Owner.Cairo, ALeft + X, ATop + Y);
  cairo_scale(Owner.Cairo, X, Y);
  cairo_arc(Owner.Cairo, 0, 0, 1, CalcAngle(X, Y, SX, SY), CalcAngle(X, Y, EX, EY));
  cairo_restore(Owner.Cairo);
  CairoStroke;
end;

procedure TCairoCanvas.Chord(x1, y1, x2, y2, StartAngle16Deg,
  EndAngle16Deg: Integer);
begin
  cairo_save(Owner.Cairo);
  cairo_translate(Owner.Cairo, x1 + (x2 - x1) / 2, y1 + (y2 - y1) / 2);
  cairo_scale(Owner.Cairo, (x2 - x1) / 2, (y2 - y1) / 2);
  cairo_arc(Owner.Cairo, 0, 0, 1, StartAngle16Deg * DegreePer16Rad, EndAngle16Deg * DegreePer16Rad);
  cairo_close_path(Owner.Cairo);
  cairo_restore(Owner.Cairo);
  CairoStroke;
  CairoFill;
end;

procedure TCairoCanvas.Chord(x1, y1, x2, y2, SX, SY, EX, EY: Integer);
var
  X, Y: Double;

begin
  X := (x2 - x1) / 2;
  Y := (y2 - y1) / 2;
  cairo_save(Owner.Cairo);
  cairo_translate(Owner.Cairo, x1 + X, y1 + Y);
  cairo_scale(Owner.Cairo, X, Y);
  cairo_arc(Owner.Cairo, 0, 0, 1, CalcAngle(X, Y, SX, SY), CalcAngle(X, Y, EX, EY));
  cairo_close_path(Owner.Cairo);
  cairo_restore(Owner.Cairo);
  CairoStroke;
  CairoFill;
end;

procedure TCairoCanvas.CopyRect(const Dest: TRect; SrcCanvas: TCanvas;
  const Source: TRect);
begin

end;

procedure TCairoCanvas.Draw(X, Y: Integer; SrcGraphic: TGraphic);
begin

end;

procedure TCairoCanvas.DrawFocusRect(const ARect: TRect);
begin

end;

procedure TCairoCanvas.StretchDraw(const DestRect: TRect; SrcGraphic: TGraphic
  );
begin

end;

procedure TCairoCanvas.Ellipse(const ARect: TRect);
begin
  Ellipse(ARect.Left, ARect.Top, ARect.Right, ARect.Bottom);
end;

procedure TCairoCanvas.Ellipse(x1, y1, x2, y2: Integer);
begin
  Arc(x1, y1, x2, y2, 0, 5760);
end;

procedure TCairoCanvas.FillRect(const ARect: TRect);
begin
  FillRect(ARect.Left, ARect.Top, ARect.Right, ARect.Bottom);
end;

procedure TCairoCanvas.FillRect(X1, Y1, X2, Y2: Integer);
begin
  cairo_rectangle(Owner.Cairo, X1, Y1, X2 - X1, Y2 - Y1);
  CairoFill;
end;

procedure TCairoCanvas.FloodFill(X, Y: Integer; FillColor: TColor;
  FillStyle: TFillStyle);
begin

end;

procedure TCairoCanvas.Frame3d(var ARect: TRect; const FrameWidth: integer;
  const Style: TGraphicsBevelCut);
begin

end;

procedure TCairoCanvas.Frame(const ARect: TRect);
begin
end;

procedure TCairoCanvas.Frame(X1, Y1, X2, Y2: Integer);
begin
end;

procedure TCairoCanvas.FrameRect(const ARect: TRect);
begin

end;

procedure TCairoCanvas.FrameRect(X1, Y1, X2, Y2: Integer);
begin

end;

procedure TCairoCanvas.GradientFill(ARect: TRect; AStart, AStop: TColor;
  ADirection: TGradientDirection);
begin

end;

procedure TCairoCanvas.Line(X1, Y1, X2, Y2: Integer);
begin
  MoveTo(X1, Y1);
  LineTo(X2, Y2);
end;

procedure TCairoCanvas.Line(const p1, p2: TPoint);
begin
  MoveTo(p1.X, p1.Y);
  LineTo(p2.X, p2.Y);
end;

procedure TCairoCanvas.Line(const Points: TRect);
begin
  MoveTo(Points.Left, Points.Top);
  LineTo(Points.Right, Points.Bottom);
end;

procedure TCairoCanvas.LineTo(X1, Y1: Integer);
begin
  cairo_line_to(Owner.Cairo, X1, Y1);
  CairoStroke;
end;

procedure TCairoCanvas.MoveTo(X1, Y1: Integer);
begin
    cairo_move_to(Owner.Cairo, X1, Y1);
end;

procedure TCairoCanvas.RadialPie(x1, y1, x2, y2, StartAngle16Deg,
  EndAngle16Deg: Integer);
begin
  cairo_save(Owner.Cairo);
  cairo_translate(Owner.Cairo, x1 + (x2 - x1) / 2, y1 + (y2 - y1) / 2);
  cairo_scale(Owner.Cairo, (x2 - x1) / 2, (y2 - y1) / 2);
  cairo_arc(Owner.Cairo, 0, 0, 1, StartAngle16Deg * DegreePer16Rad, EndAngle16Deg * DegreePer16Rad);
  cairo_line_to(Owner.Cairo, 0.5, 0.5);
  cairo_close_path(Owner.Cairo);
  cairo_restore(Owner.Cairo);
  CairoStroke;
  CairoFill;
end;

procedure TCairoCanvas.Pie(EllipseX1, EllipseY1, EllipseX2, EllipseY2, StartX,
  StartY, EndX, EndY: Integer);
var
  X, Y: Double;

begin
  X := (EllipseX2 - EllipseX1) / 2;
  Y := (EllipseY2 - EllipseY1) / 2;
  cairo_save(Owner.Cairo);
  cairo_translate(Owner.Cairo, EllipseX1 + X, EllipseY1 + Y);
  cairo_scale(Owner.Cairo, X, Y);
  cairo_arc(Owner.Cairo, 0, 0, 1, CalcAngle(X, Y, StartX, StartY), CalcAngle(X, Y, EndX, EndY));
  cairo_line_to(Owner.Cairo, 0.5, 0.5);
  cairo_close_path(Owner.Cairo);
  cairo_restore(Owner.Cairo);
  CairoStroke;
  CairoFill;
end;

procedure TCairoCanvas.PolyBezier(Points: PPoint; NumPts: Integer;
  Filled: boolean; Continuous: boolean);
begin

end;

procedure TCairoCanvas.PolyBezier(const Points: array of TPoint;
  Filled: boolean; Continuous: boolean);
begin

end;

procedure TCairoCanvas.Polygon(const Points: array of TPoint; Winding: Boolean;
  StartIndex: Integer; NumPts: Integer);
begin

end;

procedure TCairoCanvas.Polygon(Points: PPoint; NumPts: Integer; Winding: boolean
  );
begin

end;

procedure TCairoCanvas.Polygon(const Points: array of TPoint);
begin

end;

procedure TCairoCanvas.Polyline(const Points: array of TPoint;
  StartIndex: Integer; NumPts: Integer);
begin

end;

procedure TCairoCanvas.Polyline(Points: PPoint; NumPts: Integer);
begin

end;

procedure TCairoCanvas.Polyline(const Points: array of TPoint);
begin

end;

procedure TCairoCanvas.Rectangle(X1, Y1, X2, Y2: Integer);
begin
  cairo_rectangle(Owner.Cairo, X1, Y1, X2 - X1, Y2 - Y1);
  CairoStroke;
  CairoFill;
end;

procedure TCairoCanvas.Rectangle(const ARect: TRect);
begin
  Rectangle(ARect.Left, ARect.Top, ARect.Right, ARect.Bottom);
end;

procedure TCairoCanvas.RoundRect(X1, Y1, X2, Y2: Integer; RX, RY: Integer);
begin

end;

procedure TCairoCanvas.RoundRect(const Rect: TRect; RX, RY: Integer);
begin

end;

procedure TCairoCanvas.TextOut(X, Y: Integer; const Text: String);
begin
  cairo_select_font_face (Owner.Cairo, fFont.Name, fFont.Style, fFont.Weight);
  cairo_set_font_size (Owner.Cairo, fFont.Size);
  cairo_set_source_rgba(Owner.Cairo, fFont.Color.R, fFont.Color.G, fFont.Color.B, fFont.Color.A);
  cairo_move_to(Owner.Cairo, X, Y);
  cairo_show_text(Owner.Cairo, @Text[1]);
end;

procedure TCairoCanvas.TextRect(const ARect: TRect; X, Y: integer;
  const Text: string);
begin
end;

procedure TCairoCanvas.TextRect(ARect: TRect; X, Y: integer;
  const Text: string; const Style: TTextStyle);
begin

end;

function TCairoCanvas.TextExtent(const Text: string): TCairoSize;
var
  Size: TCairoSize;
  te: cairo_text_extents_t;
begin
  cairo_text_extents(Owner.Cairo, @Text[1], @te);
  Size.cx := te.width;
  Size.cy := te.height;
  Result := Size;
end;

function TCairoCanvas.TextHeight(const Text: string): Double;
begin
  Result := TextExtent(Text).cy;
end;

function TCairoCanvas.TextWidth(const Text: string): Double;
begin
  Result := TextExtent(Text).cx;
end;

{ TCairoObject }

procedure TCairoObject.CreateSurface;
begin

end;

procedure TCairoObject.CreateCairo;
begin
  fCairo := cairo_create(fSurface);
end;

constructor TCairoObject.Create;
begin
  inherited Create;
  fCanvas := TCairoCanvas.Create(Self);
end;

destructor TCairoObject.Destroy;
begin
  fCanvas.Destroy;
  inherited Destroy;
end;

procedure TCairoObject.BeginDoc;
begin
  //CreateSurface;
  CreateCairo;
end;

procedure TCairoObject.EndDoc;
begin
  cairo_Destroy(fCairo);
  cairo_surface_destroy(fSurface);
end;

procedure TCairoObject.NewPage;
begin
  cairo_show_page(Cairo);
end;


{ TCairoPrinter }
procedure TCairoPrinter.CreateSurface;
begin
  fSurface := cairo_ps_surface_create_for_stream(wf, nil, 595, 842);
end;

constructor TCairoPrinter.Create(APrinter: TPrinter; Awf: cairo_write_func_t);
begin
  inherited Create;
  fPrinter := APrinter;
  fPrinter.PrinterIndex:= 0;
  fPrinter.RawMode:= True;
  wf := Awf;
end;

procedure TCairoPrinter.BeginDoc;
begin
  CreateSurface;
  inherited BeginDoc;
  fPrinter.BeginDoc;
end;

procedure TCairoPrinter.NewPage;
begin
  inherited NewPage;
end;

procedure TCairoPrinter.EndDoc;
begin
  inherited EndDoc;
  fPrinter.EndDoc;
end;

function WritetoPrinter(closure: pointer; data:PByte; lenght: LongWord):cairo_status_t;cdecl;
  var
    i: integer;
  begin
    Printer.Write(data^, lenght, i);
    Result := CAIRO_STATUS_SUCCESS;
  end;


initialization
  Prn := TCairoPrinter.Create(Printer, @WritetoPrinter);
  
finalization
  Prn.Destroy;

end.

